base_options: dict = {
    "style": None,
    "sectional_overwrite": [],
    "sectional_overwrite_no_negate": [],
    "ordering": [],
    "indent_adjust": [],
    "parent_allows_duplicate_child": [],
    "sectional_exiting": [],
    "full_text_sub": [],
    "per_line_sub": [],
    "idempotent_commands_blacklist": [],
    "idempotent_commands": [],
    "negation_default_when": [],
    "negation_negate_with": [],
}
ios_options: dict = {
    "style": "ios",
    "ordering": [
        {"lineage": [{"startswith": "no vlan filter"}], "order": 700},
        {
            "lineage": [
                {"startswith": "interface"},
                {"startswith": "no shutdown"},
            ],
            "order": 700,
        },
    ],
    "sectional_exiting": [
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "template peer-policy"},
            ],
            "exit_text": "exit-peer-policy",
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "template peer-session"},
            ],
            "exit_text": "exit-peer-session",
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "address-family"},
            ],
            "exit_text": "exit-address-family",
        },
    ],
    "per_line_sub": [
        {"search": "^Building configuration.*", "replace": ""},
        {"search": "^Current configuration.*", "replace": ""},
        {"search": "^! Last configuration change.*", "replace": ""},
        {"search": "^! NVRAM config last updated.*", "replace": ""},
        {"search": "^ntp clock-period .*", "replace": ""},
        {"search": "^version.*", "replace": ""},
        {"search": "^ logging event link-status$", "replace": ""},
        {"search": "^ logging event subif-link-status$", "replace": ""},
        {"search": "^\\s*ipv6 unreachables disable$", "replace": ""},
        {"search": "^end$", "replace": ""},
        {"search": "^\\s*[#!].*", "replace": ""},
        {"search": "^ no ip address", "replace": ""},
        {"search": "^ exit-peer-policy", "replace": ""},
        {"search": "^ exit-peer-session", "replace": ""},
        {"search": "^ exit-address-family", "replace": ""},
        {"search": "^crypto key generate rsa general-keys.*$", "replace": ""},
    ],
    "idempotent_commands": [
        {"lineage": [{"startswith": "vlan"}, {"startswith": "name"}]},
        {
            "lineage": [
                {"startswith": "interface"},
                {"startswith": ["description", "ip address"]},
            ]
        },
    ],
}
iosxe_options: dict = {
    "style": "ios",
    "sectional_overwrite": [{"lineage": [{"startswith": "ipv6 access-list"}]}],
    "sectional_exiting": [
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "template peer-policy"},
            ],
            "exit_text": "exit-peer-policy",
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "template peer-session"},
            ],
            "exit_text": "exit-peer-session",
        },
        {
            "lineage": [{"startswith": "router bgp"}, {"startswith": "address-family"}],
            "exit_text": "exit-address-family",
        },
    ],
    "per_line_sub": [
        {"search": "^Building configuration.*", "replace": ""},
        {"search": "^Current configuration.*", "replace": ""},
        {"search": "^! Last configuration change.*", "replace": ""},
        {"search": "^! NVRAM config last updated.*", "replace": ""},
        {"search": "^ntp clock-period .*", "replace": ""},
        {"search": "^version.*", "replace": ""},
        {"search": "^ logging event link-status$", "replace": ""},
        {"search": "^ logging event subif-link-status$", "replace": ""},
        {"search": "^\\s*ipv6 unreachables disable$", "replace": ""},
        {"search": "^end$", "replace": ""},
        {"search": "^ no ip address", "replace": ""},
        {"search": "^ exit-peer-policy", "replace": ""},
        {"search": "^ exit-peer-session", "replace": ""},
        {"search": "^ exit-address-family", "replace": ""},
    ],
    "idempotent_commands": [
        {
            "lineage": [
                {"startswith": "router ospf"},
                {"startswith": ["log-adjacency-changes"]},
            ]
        },
        {"lineage": [{"startswith": "router ospf"}, {"startswith": ["router-id"]}]},
        {
            "lineage": [
                {"startswith": "ipv6 router ospf"},
                {"startswith": ["log-adjacency-changes"]},
            ]
        },
        {
            "lineage": [
                {"startswith": "ipv6 router ospf"},
                {"startswith": ["router-id"]},
            ]
        },
        {"lineage": [{"startswith": "router bgp"}, {"startswith": "bgp router-id"}]},
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"re_search": "neighbor \\S+ description"},
            ]
        },
        {"lineage": [{"startswith": ["hostname"]}]},
        {"lineage": [{"contains": ["source-interface", "trap-source"]}]},
        {"lineage": [{"startswith": ["snmp-server community"]}]},
        {"lineage": [{"startswith": ["mac address-table aging-time"]}]},
        {"lineage": [{"startswith": ["aaa authentication"]}]},
        {"lineage": [{"startswith": ["aaa authorization"]}]},
        {"lineage": [{"startswith": ["errdisable recovery"]}]},
        {"lineage": [{"startswith": "line"}, {"startswith": ["access-class"]}]},
        {"lineage": [{"startswith": "line"}, {"startswith": ["ipv6 access-class"]}]},
        {"lineage": [{"startswith": "interface"}, {"startswith": ["ip ospf cost"]}]},
        {"lineage": [{"startswith": "interface"}, {"startswith": ["ipv6 ospf cost"]}]},
        {
            "lineage": [
                {"startswith": "interface"},
                {"re_search": ["standby \\d authentication"]},
            ]
        },
        {
            "lineage": [
                {"startswith": "interface"},
                {"re_search": ["standby \\d priority"]},
            ]
        },
        {"lineage": [{"startswith": "username admin "}]},
        {
            "lineage": [
                {"startswith": "policy-map system-cpp-policy"},
                {"startswith": "class"},
                {"startswith": "police"},
            ]
        },
        {"lineage": [{"startswith": "banner"}]},
        {"lineage": [{"startswith": "logging facility"}]},
        {"lineage": [{"startswith": "ip tftp source-interface"}]},
        {"lineage": [{"startswith": "snmp-server trap-source"}]},
        {"lineage": [{"startswith": "power redundancy-mode"}]},
    ],
}
iosxr_options: dict = {
    "style": "iosxr",
    "ordering": [
        {"lineage": [{"startswith": "vrf "}], "order": 300},
        {"lineage": [{"startswith": "no vrf "}], "order": 700},
    ],
    "sectional_overwrite": [{"lineage": [{"startswith": "template"}]}],
    "sectional_overwrite_no_negate": [
        {"lineage": [{"startswith": "as-path-set"}]},
        {"lineage": [{"startswith": "prefix-set"}]},
        {"lineage": [{"startswith": "route-policy"}]},
        {"lineage": [{"startswith": "extcommunity-set"}]},
        {"lineage": [{"startswith": "community-set"}]},
    ],
    "parent_allows_duplicate_child": [{"lineage": [{"startswith": "route-policy"}]}],
    "sectional_exiting": [
        {"lineage": [{"startswith": "route-policy"}], "exit_text": "end-policy"},
        {"lineage": [{"startswith": "prefix-set"}], "exit_text": "end-set"},
        {"lineage": [{"startswith": "policy-map"}], "exit_text": "end-policy-map"},
        {"lineage": [{"startswith": "class-map"}], "exit_text": "end-class-map"},
        {"lineage": [{"startswith": "community-set"}], "exit_text": "end-set"},
        {"lineage": [{"startswith": "extcommunity-set"}], "exit_text": "end-set"},
        {"lineage": [{"equals": "rsvp"}], "exit_text": "exit"},
        {"lineage": [{"equals": "mpls traffic-eng"}], "exit_text": "exit"},
        {"lineage": [{"startswith": "mpls ldp"}], "exit_text": "exit"},
        {"lineage": [{"startswith": "router ospf"}], "exit_text": "exit"},
        {"lineage": [{"startswith": "router ospfv3"}], "exit_text": "exit"},
        {"lineage": [{"startswith": "template"}], "exit_text": "end-template"},
        {"lineage": [{"startswith": "interface"}], "exit_text": "root"},
        {"lineage": [{"startswith": "router bgp"}], "exit_text": "root"},
    ],
    "indent_adjust": [
        {"start_expression": "^\\s*template", "end_expression": "^\\s*end-template"}
    ],
    "per_line_sub": [
        {"search": "^Building configuration.*", "replace": ""},
        {"search": "^Current configuration.*", "replace": ""},
        {"search": "^ntp clock-period .*", "replace": ""},
        {"search": ".*speed.*", "replace": ""},
        {"search": ".*duplex.*", "replace": ""},
        {"search": ".*negotiation auto.*", "replace": ""},
        {"search": ".*parity none.*", "replace": ""},
        {"search": "^end-policy$", "replace": " end-policy"},
        {"search": "^end-set$", "replace": " end-set"},
        {"search": "^end$", "replace": ""},
        {"search": "^\\s*[#!].*", "replace": ""},
    ],
    "idempotent_commands": [
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "vrf"},
                {"startswith": "address-family"},
                {"startswith": "additional-paths selection route-policy"},
            ]
        },
        {"lineage": [{"startswith": "router bgp"}, {"startswith": "bgp router-id"}]},
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "neighbor-group"},
                {"startswith": "address-family"},
                {"startswith": "soft-reconfiguration inbound"},
            ]
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "vrf"},
                {"startswith": "neighbor"},
                {"startswith": "address-family"},
                {"startswith": ["soft-reconfiguration inbound", "maximum-prefix"]},
            ]
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "vrf"},
                {"startswith": "neighbor"},
                {"startswith": ["password", "description"]},
            ]
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "neighbor"},
                {"startswith": ["description", "password"]},
            ]
        },
        {
            "lineage": [
                {"startswith": "router ospf"},
                {"startswith": "area"},
                {"startswith": "interface"},
                {"startswith": "cost"},
            ]
        },
        {"lineage": [{"startswith": "router ospf"}, {"startswith": "router-id"}]},
        {
            "lineage": [
                {"startswith": "router ospf"},
                {"startswith": "area"},
                {"startswith": "message-digest-key"},
            ]
        },
        {
            "lineage": [
                {"startswith": "router ospf"},
                {"startswith": "max-metric router-lsa"},
            ]
        },
        {"lineage": [{"equals": "l2vpn"}, {"startswith": "router-id"}]},
        {"lineage": [{"re_search": "logging \\d+.\\d+.\\d+.\\d+ vrf MGMT"}]},
        {
            "lineage": [
                {"equals": "line default"},
                {"startswith": "access-class ingress"},
            ]
        },
        {"lineage": [{"equals": "line default"}, {"startswith": "transport input"}]},
        {"lineage": [{"startswith": "hostname"}]},
        {"lineage": [{"startswith": "logging source-interface"}]},
        {"lineage": [{"startswith": "interface"}, {"startswith": "ipv4 address"}]},
        {"lineage": [{"startswith": "snmp-server community"}]},
        {"lineage": [{"startswith": "snmp-server location"}]},
        {"lineage": [{"equals": "line console"}, {"startswith": "exec-timeout"}]},
        {
            "lineage": [
                {"equals": "mpls ldp"},
                {"startswith": "session protection duration"},
            ]
        },
        {"lineage": [{"equals": "mpls ldp"}, {"startswith": "igp sync delay"}]},
        {"lineage": [{"startswith": "interface"}, {"startswith": ["mtu"]}]},
        {"lineage": [{"startswith": "banner"}]},
    ],
}
nxos_options: dict = {
    "style": "nxos",
    "per_line_sub": [
        {"search": "^Building configuration.*", "replace": ""},
        {"search": "^Current configuration.*", "replace": ""},
        {"search": "^ntp clock-period .*", "replace": ""},
        {"search": "^snmp-server location  ", "replace": "snmp-server location "},
        {"search": "^version.*", "replace": ""},
        {"search": "^boot (system|kickstart) .*", "replace": ""},
        {"search": "!.*", "replace": ""},
    ],
    "idempotent_commands_blacklist": [
        {
            "lineage": [
                {"startswith": "interface"},
                {"re_search": "ip address.*secondary"},
            ]
        }
    ],
    "idempotent_commands": [
        {
            "lineage": [
                {
                    "startswith": [
                        "power redundancy-mode",
                        "cli alias name wr ",
                        "aaa authentication login console",
                        "port-channel load-balance",
                        "hostname",
                        "ip tftp source-interface",
                        "ip telnet source-interface",
                        "ip tacacs source-interface",
                        "logging source-interface",
                        "tacacs-server",
                    ],
                    "re_search": "^spanning-tree vlan ([\\d,-]+) priority",
                }
            ]
        },
        {"lineage": [{"startswith": ["hardware access-list tcam region ifacl"]}]},
        {"lineage": [{"startswith": ["hardware access-list tcam region vacl"]}]},
        {"lineage": [{"startswith": ["hardware access-list tcam region qos"]}]},
        {"lineage": [{"startswith": ["hardware access-list tcam region racl"]}]},
        {"lineage": [{"startswith": ["hardware access-list tcam region ipv6-racl"]}]},
        {"lineage": [{"startswith": ["hardware access-list tcam region e-ipv6-racl"]}]},
        {"lineage": [{"startswith": ["hardware access-list tcam region l3qos"]}]},
        {
            "lineage": [
                {"startswith": "router ospf"},
                {"startswith": "vrf"},
                {"startswith": ["maximum-paths", "log-adjacency-changes"]},
            ]
        },
        {
            "lineage": [
                {"startswith": "router ospf"},
                {"startswith": ["maximum-paths", "log-adjacency-changes"]},
            ]
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "vrf"},
                {"startswith": "address-family"},
                {"startswith": ["maximum-paths"]},
            ]
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "address-family"},
                {"startswith": ["maximum-paths"]},
            ]
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "template"},
                {"startswith": "address-family"},
                {"startswith": "send-community"},
            ]
        },
        {
            "lineage": [
                {"startswith": "interface"},
                {"re_search": "^hsrp \\d+"},
                {"startswith": ["ip", "priority", "authentication md5 key-string"]},
            ]
        },
        {
            "lineage": [
                {"startswith": "interface"},
                {
                    "startswith": [
                        "ip address",
                        "duplex",
                        "speed",
                        "switchport mode",
                        "switchport access vlan",
                        "switchport trunk native vlan",
                        "switchport trunk allowed vlan",
                        "udld port",
                        "ip ospf cost",
                        "ipv6 link-local",
                        "ospfv3 cost",
                    ]
                },
            ]
        },
        {"lineage": [{"startswith": "interface"}, {"startswith": "mtu"}]},
        {"lineage": [{"equals": "line console"}, {"startswith": "exec-timeout"}]},
        {
            "lineage": [
                {"startswith": "line vty"},
                {
                    "startswith": [
                        "transport input",
                        "ipv6 access-class",
                        "access-class",
                    ]
                },
            ]
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {
                    "startswith": "bgp router-id",
                    "re_search": "neighbor \\S+ description",
                },
            ]
        },
        {
            "lineage": [
                {"startswith": "router ospf"},
                {"startswith": ["router-id", "log-adjacency-changes"]},
            ]
        },
        {
            "lineage": [
                {"startswith": "ipv6 router ospf"},
                {"startswith": ["router-id", "log-adjacency-changes"]},
            ]
        },
        {
            "lineage": [
                {
                    "startswith": [
                        "mac address-table aging-time",
                        "snmp-server community",
                        "snmp-server location",
                    ]
                }
            ]
        },
        {"lineage": [{"startswith": "vpc domain"}, {"startswith": "role priority"}]},
        {"lineage": [{"startswith": "banner"}]},
        {"lineage": [{"startswith": "username admin password 5"}]},
        {
            "lineage": [
                {"equals": "policy-map type control-plane copp-system-policy"},
                {"startswith": "class"},
                {"startswith": "police"},
            ]
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "vrf"},
                {"startswith": "neighbor"},
                {"startswith": "address-family"},
                {"startswith": "soft-reconfiguration inbound"},
            ]
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "vrf"},
                {"startswith": "neighbor"},
                {"startswith": "password"},
            ]
        },
    ],
    "negation_default_when": [
        {
            "lineage": [
                {"startswith": "interface"},
                {
                    "startswith": "ip ospf bfd",
                    "re_search": "standby \\d+ authentication md5 key-string",
                },
            ]
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "neighbor"},
                {"startswith": "address-family"},
                {"equals": "send-community"},
            ]
        },
        {
            "lineage": [
                {"startswith": "interface"},
                {"contains": "ip ospf passive-interface"},
            ]
        },
        {
            "lineage": [
                {"startswith": "interface"},
                {"contains": "ospfv3 passive-interface"},
            ]
        },
    ],
    "negation_negate_with": [
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "address-family"},
                {"startswith": "maximum-paths ibgp"},
            ],
            "use": "default maximum-paths ibgp",
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "vrf"},
                {"startswith": "address-family"},
                {"startswith": "maximum-paths ibgp"},
            ],
            "use": "default maximum-paths ibgp",
        },
        {
            "lineage": [{"equals": "line vty"}, {"startswith": "session-limit"}],
            "use": "session-limit 32",
        },
    ],
}
eos_options: dict = {
    "style": "eos",
    "sectional_exiting": [
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "template peer-policy"},
            ],
            "exit_text": "exit-peer-policy",
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"startswith": "template peer-session"},
            ],
            "exit_text": "exit-peer-session",
        },
        {
            "lineage": [{"startswith": "router bgp"}, {"startswith": "address-family"}],
            "exit_text": "exit-address-family",
        },
    ],
    "per_line_sub": [
        {"search": "^Building configuration.*", "replace": ""},
        {"search": "^Current configuration.*", "replace": ""},
        {"search": "^! Last configuration change.*", "replace": ""},
        {"search": "^! NVRAM config last updated.*", "replace": ""},
        {"search": "^ntp clock-period .*", "replace": ""},
        {"search": "^version.*", "replace": ""},
        {"search": "^ logging event link-status$", "replace": ""},
        {"search": "^ logging event subif-link-status$", "replace": ""},
        {"search": "^\\s*ipv6 unreachables disable$", "replace": ""},
        {"search": "^end$", "replace": ""},
        {"search": "^\\s*[#!].*", "replace": ""},
        {"search": "^ no ip address", "replace": ""},
        {"search": "^ exit-peer-policy", "replace": ""},
        {"search": "^ exit-peer-session", "replace": ""},
        {"search": "^ exit-address-family", "replace": ""},
    ],
    "idempotent_commands": [
        {"lineage": [{"startswith": "hostname"}]},
        {"lineage": [{"startswith": "logging source-interface"}]},
        {"lineage": [{"startswith": "interface"}, {"startswith": "ip address"}]},
        {
            "lineage": [
                {"startswith": "line vty"},
                {
                    "startswith": [
                        "transport input",
                        "access-class",
                        "ipv6 access-class",
                    ]
                },
            ]
        },
        {
            "lineage": [
                {"startswith": "interface"},
                {"re_search": "standby \\d+ (priority|authentication md5)"},
            ]
        },
        {"lineage": [{"startswith": "router bgp"}, {"startswith": "bgp router-id"}]},
        {
            "lineage": [
                {"startswith": "router ospf"},
                {"startswith": ["router-id", "max-lsa", "maximum-paths"]},
            ]
        },
        {"lineage": [{"startswith": "ipv6 router ospf"}, {"startswith": "router-id"}]},
        {
            "lineage": [
                {"startswith": "router ospf"},
                {"startswith": "log-adjacency-changes"},
            ]
        },
        {
            "lineage": [
                {"startswith": "ipv6 router ospf"},
                {"startswith": "log-adjacency-changes"},
            ]
        },
        {
            "lineage": [
                {"startswith": "router bgp"},
                {"re_search": "neighbor \\S+ description"},
            ]
        },
        {"lineage": [{"startswith": "snmp-server community"}]},
        {"lineage": [{"startswith": "snmp-server location"}]},
        {"lineage": [{"equals": "line con 0"}, {"startswith": "exec-timeout"}]},
        {
            "lineage": [
                {"startswith": "interface"},
                {"startswith": "ip ospf message-digest-key"},
            ]
        },
        {"lineage": [{"startswith": "logging buffered"}]},
        {"lineage": [{"startswith": "tacacs-server key"}]},
        {"lineage": [{"startswith": "logging facility"}]},
        {"lineage": [{"startswith": "vlan internal allocation policy"}]},
        {"lineage": [{"startswith": "username admin"}]},
        {"lineage": [{"startswith": "snmp-server user"}]},
        {"lineage": [{"startswith": "banner"}]},
        {"lineage": [{"startswith": "ntp source"}]},
        {"lineage": [{"startswith": "management"}, {"startswith": "idle-timeout"}]},
        {
            "lineage": [
                {"startswith": "aaa authentication enable default group tacacs+"}
            ]
        },
        {
            "lineage": [
                {"equals": "control-plane"},
                {"equals": "ip access-group CPP in"},
            ]
        },
        {"lineage": [{"startswith": "interface"}, {"startswith": "mtu"}]},
        {"lineage": [{"startswith": "snmp-server source-interface"}]},
        {"lineage": [{"startswith": "ip tftp client source-interface"}]},
    ],
    "negation_default_when": [
        {
            "lineage": [
                {"startswith": "interface"},
                {"equals": "logging event link-status"},
            ]
        }
    ],
}


def options_for(os: str) -> dict:
    """ Create base options on an OS level. """
    options: dict = {
        "ios": ios_options,
        "iosxe": iosxe_options,
        "iosxr": iosxr_options,
        "nxos": nxos_options,
        "eos": eos_options,
    }

    if options.get(os):
        return {**base_options, **options[os]}

    return {**base_options, "style": os}



from __future__ import annotations
from typing import (
    Optional,
    List,
    Iterator,
    Dict,
    Tuple,
    Union,
    Set,
    TYPE_CHECKING,
    Type,
)
from logging import getLogger
from abc import ABC, abstractmethod
from functools import cached_property

from . import text_match

if TYPE_CHECKING:
    from .child import HConfigChild
    from .root import HConfig
    from .host import Host

logger = getLogger(__name__)


class HConfigBase(ABC):  # pylint: disable=too-many-public-methods
    def __init__(self) -> None:
        self.children: List[HConfigChild] = []
        self.children_dict: Dict[str, HConfigChild] = {}
        self.host: Host

    def __str__(self) -> str:
        return "\n".join(c.cisco_style_text() for c in self.all_children())

    def __len__(self) -> int:
        return len(list(self.all_children()))

    def __bool__(self) -> bool:
        return bool(self.children)

    def __contains__(self, item: str) -> bool:
        return item in self.children_dict

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HConfigBase):
            return NotImplemented

        if len(self.children) != len(other.children):
            return False

        for self_child, other_child in zip(
            sorted(self.children), sorted(other.children)
        ):
            if self_child != other_child:
                return False

        return True

    @abstractmethod
    def _duplicate_child_allowed_check(self) -> bool:
        pass

    @property
    @abstractmethod
    def options(self) -> dict:
        pass

    @property
    @abstractmethod
    def root(self) -> HConfig:
        pass

    @abstractmethod
    def lineage(self) -> Iterator[HConfigChild]:
        pass

    @abstractmethod
    def depth(self) -> int:
        pass

    @property
    @abstractmethod
    def logs(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def _child_class(self) -> Type[HConfigChild]:
        pass

    def has_children(self) -> bool:
        return bool(self.children)

    def add_children(self, lines: List[str]) -> None:
        """
        Add child instances of HConfigChild

        :param lines: HConfigChild object -> type list
        :return: None
        """
        for line in lines:
            self.add_child(line)

    def add_child(
        self,
        text: str,
        alert_on_duplicate: bool = False,
        idx: Optional[int] = None,
        force_duplicate: bool = False,
    ) -> HConfigChild:
        """ Add a child instance of HConfigChild """

        if idx is None:
            idx = len(self.children)
        # if child does not exist
        if text not in self:
            new_item = self._child_class(self, text)  # type: ignore
            self.children.insert(idx, new_item)
            self.children_dict[text] = new_item
            return new_item
        # if child does exist and is allowed to be installed as a duplicate
        if self._duplicate_child_allowed_check() or force_duplicate:
            new_item = self._child_class(self, text)  # type: ignore
            self.children.insert(idx, new_item)
            self.rebuild_children_dict()
            return new_item

        # If the child is already present and the parent does not allow
        # duplicate children, return the existing child
        # Ignore duplicate remarks in ACLs
        if alert_on_duplicate and not text.startswith("remark "):
            self.logs.append(f"Found a duplicate section: {list(self.path()) + [text]}")
        return self.children_dict[text]

    def path(self) -> Iterator[str]:  # pylint: disable=no-self-use
        yield from ()

    def add_deep_copy_of(
        self, child_to_add: HConfigChild, merged: bool = False
    ) -> HConfigChild:
        """Add a nested copy of a child to self"""
        new_child = self.add_shallow_copy_of(child_to_add, merged=merged)
        for child in child_to_add.children:
            new_child.add_deep_copy_of(child, merged=merged)

        return new_child

    def to_tag_spec(self, tags: Set[str]) -> List[dict]:
        """
        Returns the configuration as a tag spec definition

        This is handy when you have a segment of config and need to
        generate a tag spec to tag configuration in another instance
        """
        tag_spec = []
        for child in self.all_children():
            if not child.children:
                child_spec = [{"equals": t} for t in child.path()]
                tag_spec.append({"section": child_spec, "add_tags": tags})
        return tag_spec

    def del_child_by_text(self, text: str) -> None:
        """ Delete all children with the provided text """
        if text in self.children_dict:
            self.children[:] = [c for c in self.children if c.text != text]
            self.rebuild_children_dict()

    def del_child(self, child: HConfigChild) -> None:
        """Delete a child from self.children and self.children_dict"""
        try:
            self.children.remove(child)
        except ValueError:
            pass
        else:
            self.rebuild_children_dict()

    def all_children_sorted_untagged(self) -> Iterator[HConfigChild]:
        """ Yield all children recursively that are untagged """
        yield from (c for c in self.all_children_sorted() if None in c.tags)

    def all_children_sorted_by_tags(
        self, include_tags: Set[str], exclude_tags: Set[str]
    ) -> Iterator[HConfigChild]:
        """ Yield all children recursively that match include/exlcude tags """
        for child in self.all_children_sorted():
            if child.line_inclusion_test(include_tags, exclude_tags):
                yield child

    def all_children_sorted(self) -> Iterator[HConfigChild]:
        """ Recursively find and yield all children sorted at each hierarchy """
        for child in sorted(self.children):
            yield child
            yield from child.all_children_sorted()

    def all_children_sorted_with_lineage_rules(
        self, rules: List[dict]
    ) -> Iterator[HConfigChild]:
        """ Recursively find and yield all children sorted at each hierarchy given lineage rules """
        yielded = set()
        matched: Set[HConfigChild] = set()
        # pylint: disable=too-many-nested-blocks
        for child in self.all_children_sorted():
            for ancestor in child.lineage():
                if ancestor in matched:
                    yield child
                    yielded.add(child)
                    break
            else:
                for rule in rules:
                    if child.lineage_test(rule, False):
                        matched.add(child)
                        for ancestor in child.lineage():
                            if ancestor in yielded:
                                continue
                            yield ancestor
                            yielded.add(ancestor)
                        break

    def all_children(self) -> Iterator[HConfigChild]:
        """ Recursively find and yield all children at each hierarchy """
        for child in self.children:
            yield child
            yield from child.all_children()

    def get_child(self, test: str, expression: str) -> Optional[HConfigChild]:
        """ Find a child by text_match rule. If it is not found, return None """
        if test == "equals":
            return self.children_dict.get(expression, None)

        try:
            return next(self.get_children(test, expression))
        except StopIteration:
            return None

    def get_child_deep(
        self, test_expression_pairs: List[Tuple[str, str]]
    ) -> Optional[HConfigChild]:
        """
        Find a child recursively with a list of test/expression pairs

        e.g.

        .. code:: python

            result = hier_obj.get_child_deep([('equals', 'control-plane'),
                                              ('equals', 'service-policy input system-cpp-policy')])
        """

        test, expression = test_expression_pairs.pop(0)
        if test == "equals":
            result = self.children_dict.get(expression, None)
            if result and test_expression_pairs:
                return result.get_child_deep(test_expression_pairs)
            return result

        try:
            result = next(self.get_children(test, expression))
        except StopIteration:
            return None
        if result and test_expression_pairs:
            return result.get_child_deep(test_expression_pairs)
        return result

    def get_children(self, test: str, expression: str) -> Iterator[HConfigChild]:
        """ Find all children matching a text_match rule and return them. """
        for child in self.children:
            if text_match.dict_call(test, child.text, expression):
                yield child

    def add_shallow_copy_of(
        self, child_to_add: HConfigChild, merged: bool = False
    ) -> HConfigChild:
        """ Add a nested copy of a child_to_add to self.children """

        new_child = self.add_child(child_to_add.text)

        if merged:
            new_child.instances.append(
                {
                    "hostname": child_to_add.host.hostname,
                    "comments": child_to_add.comments,
                    "tags": child_to_add.tags,
                }
            )
        new_child.comments.update(child_to_add.comments)
        new_child.tags.update(child_to_add.tags)
        new_child.order_weight = child_to_add.order_weight

        return new_child

    def rebuild_children_dict(self) -> None:
        """ Rebuild self.children_dict """
        self.children_dict = {}
        for child in self.children:
            self.children_dict.setdefault(child.text, child)

    def delete_all_children(self) -> None:
        """ Delete all children """
        self.children.clear()
        self.rebuild_children_dict()

    def _with_tags(
        self, tags: Set[str], new_instance: Union[HConfig, HConfigChild]
    ) -> Union[HConfig, HConfigChild]:
        """
        Returns a new instance containing only sub-objects
        with one of the tags in tags
        """
        for child in self.children:
            if tags.intersection(child.tags):
                new_child = new_instance.add_shallow_copy_of(child)
                # pylint: disable=protected-access
                child._with_tags(tags, new_instance=new_child)

        return new_instance

    def _config_to_get_to(
        self, target: Union[HConfig, HConfigChild], delta: Union[HConfig, HConfigChild]
    ) -> Union[HConfig, HConfigChild]:
        """
        Figures out what commands need to be executed to transition from self to target.
        self is the source data structure(i.e. the running_config),
        target is the destination(i.e. generated_config)

        """
        self._config_to_get_to_left(target, delta)
        self._config_to_get_to_right(target, delta)

        return delta

    def _difference(
        self,
        target: Union[HConfig, HConfigChild],
        delta: Union[HConfig, HConfigChild],
    ) -> Union[HConfig, HConfigChild]:
        for self_child in self.children:
            # Not dealing with negations and defaults for now
            if self_child.text.startswith((self._negation_prefix, "default ")):
                continue

            target_child = target.get_child("equals", self_child.text)

            if target_child is None:
                delta.add_deep_copy_of(self_child)
            else:
                delta_child = delta.add_child(self_child.text)
                # pylint: disable=protected-access
                self_child._difference(target_child, delta_child)
                if not delta_child.children:
                    delta_child.delete()

        return delta

    def _config_to_get_to_left(
        self, target: Union[HConfig, HConfigChild], delta: Union[HConfig, HConfigChild]
    ) -> None:
        # find self.children that are not in target.children
        # i.e. what needs to be negated or defaulted
        # Also, find out if another command in self.children will overwrite
        # i.e. be idempotent
        for self_child in self.children:
            if self_child.text in target:
                continue
            if self_child.is_idempotent_command(target.children):
                continue

            # in other but not self
            # add this node but not any children
            deleted = delta.add_child(self_child.text)
            deleted.negate()
            if self_child.children:
                deleted.comments.add(f"removes {len(self_child.children) + 1} lines")

    def _config_to_get_to_right(
        self, target: Union[HConfig, HConfigChild], delta: Union[HConfig, HConfigChild]
    ) -> None:
        # find what would need to be added to source_config to get to self
        for target_child in target.children:
            # if the child exist, recurse into its children
            self_child = self.get_child("equals", target_child.text)
            # the child is absent, add it
            if self_child is None:
                new_item = delta.add_deep_copy_of(target_child)
                # mark the new item and all of its children as new_in_config
                new_item.new_in_config = True
                for child in new_item.all_children():
                    child.new_in_config = True
                if new_item.children:
                    new_item.comments.add("new section")
            else:
                # This creates a new HConfigChild object just in case there are some delta children
                # Not very efficient, think of a way to not do this
                subtree = delta.add_child(target_child.text)
                # pylint: disable=protected-access
                self_child._config_to_get_to(target_child, subtree)
                if not subtree.children:
                    subtree.delete()
                # Do we need to rewrite the child and its children as well?
                elif self_child.sectional_overwrite_check():
                    target_child.overwrite_with(self_child, delta, True)
                elif self_child.sectional_overwrite_no_negate_check():
                    target_child.overwrite_with(self_child, delta, False)

    @cached_property
    def _negation_prefix(self) -> str:
        return str(self.options["negation"]) + " "



from __future__ import annotations
from typing import (
    Optional,
    Set,
    Union,
    Iterator,
    List,
    TYPE_CHECKING,
    Type,
    Iterable,
    Tuple,
)
from logging import getLogger

from .base import HConfigBase
from . import text_match

if TYPE_CHECKING:
    from .root import HConfig


logger = getLogger(__name__)


# pylint: disable=too-many-instance-attributes,too-many-public-methods
class HConfigChild(HConfigBase):
    def __init__(self, parent: Union[HConfig, HConfigChild], text: str):
        super().__init__()
        self.parent = parent
        self.host = self.root.host
        self._text: str = text.strip()
        self.real_indent_level: int
        # The intent is for self.order_weight values to range from 1 to 999
        # with the default weight being 500
        self.order_weight: int = 500
        self._tags: Set[str] = set()
        self.comments: Set[str] = set()
        self.new_in_config: bool = False
        self.instances: List[dict] = []
        self.facts: dict = {}  # To store externally inserted facts

    def __repr__(self) -> str:
        return f"HConfigChild(HConfig{'' if self.parent is self.root else 'Child'}, {self.text})"

    def __lt__(self, other: HConfigChild) -> bool:
        return self.order_weight < other.order_weight

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HConfigChild):
            return NotImplemented

        if (
            self.text != other.text
            or self.tags != other.tags
            or self.comments != other.comments
            or self.new_in_config != other.new_in_config
        ):
            return False
        return super().__eq__(other)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, value: str) -> None:
        """
        Used for when self.text is changed after the object
        is instantiated to rebuild the children dictionary
        """
        self._text = value.strip()
        self.parent.rebuild_children_dict()

    @property
    def root(self) -> HConfig:
        """ returns the HConfig object at the base of the tree """
        return self.parent.root

    @property
    def logs(self) -> List[str]:
        return self.root.logs

    @property
    def options(self) -> dict:
        return self.root.options

    @property
    def _child_class(self) -> Type[HConfigChild]:
        return HConfigChild

    def depth(self) -> int:
        """ Returns the distance to the root HConfig object i.e. indent level """
        return self.parent.depth() + 1

    # Consider removing, no internal or other use that I am aware of
    def move(self, new_parent: Union[HConfig, HConfigChild]) -> None:
        """
        move one HConfigChild object to different HConfig parent object

        .. code:: python

            hier1 = HConfig(host=host)
            interface1 = hier1.add_child('interface Vlan2')
            interface1.add_child('ip address 10.0.0.1 255.255.255.252')

            hier2 = Hconfig(host)

            interface1.move(hier2)

        :param new_parent: HConfigChild object -> type list
        :return: None
        """
        new_parent.children.append(self)
        new_parent.rebuild_children_dict()
        self.delete()

    def lineage(self) -> Iterator[HConfigChild]:
        """ Yields the lineage of parent objects, up to but excluding the root """
        yield from self.parent.lineage()
        yield self

    def path(self) -> Iterator[str]:
        """ Return a list of the text instance variables from self.lineage """
        for parent in self.lineage():
            yield parent.text

    def cisco_style_text(
        self, style: str = "without_comments", tag: Optional[str] = None
    ) -> str:
        """ Return a Cisco style formated line i.e. indentation_level + text ! comments """

        comments = []
        if style == "without_comments":
            pass
        elif style == "merged":
            # count the number of instances that have the tag
            instance_count = 0
            instance_comments: Set[str] = set()
            for instance in self.instances:
                if tag is None or tag in instance["tags"]:
                    instance_count += 1
                    instance_comments.update(instance["comments"])

            # should the word 'instance' be plural?
            word = "instance" if instance_count == 1 else "instances"

            comments.append(f"{instance_count} {word}")
            comments.extend(instance_comments)
        elif style == "with_comments":
            comments.extend(self.comments)

        return "{}{}{}".format(
            "  " * (self.depth() - 1),  # render the indentation
            self.text,
            " !{}".format(", ".join(sorted(comments))) if comments else "",
        )

    def delete(self) -> None:
        """ Delete the current object from its parent """
        self.parent.del_child(self)

    def append_tag(self, tag: str) -> None:
        """
        Add a tag to self._tags on all leaf nodes
        """
        if self.is_branch:
            for child in self.children:
                child.append_tag(tag)
        else:
            self._tags.add(tag)

    def append_tags(self, tags: Union[str, List[str], Set[str]]) -> None:
        """
        Add tags to self._tags on all leaf nodes
        """
        tags = self._to_set(tags)
        if self.is_branch:
            for child in self.children:
                child.append_tags(tags)
        else:
            self._tags.update(tags)

    def remove_tag(self, tag: str) -> None:
        """
        Remove a tag from self._tags on all leaf nodes
        """
        if self.is_branch:
            for child in self.children:
                child.remove_tag(tag)
        else:
            self._tags.remove(tag)

    def remove_tags(self, tags: Union[str, List[str], Set[str]]) -> None:
        """
        Remove tags from self._tags on all leaf nodes
        """
        tags = self._to_set(tags)
        if self.is_branch:
            for child in self.children:
                child.remove_tags(tags)
        else:
            self._tags.difference_update(tags)

    def negate(self) -> HConfigChild:
        """ Negate self.text """
        for rule in self.options["negation_negate_with"]:
            if self.lineage_test(rule):
                self.text = rule["use"]
                return self

        for rule in self.options["negation_default_when"]:
            if self.lineage_test(rule):
                return self._default()

        return self._swap_negation()

    @property
    def is_leaf(self) -> bool:
        """ returns True if there are no children and is not an instance of HConfig """
        return not self.is_branch

    @property
    def is_branch(self) -> bool:
        """ returns True if there are children or is an instance of HConfig """
        return bool(self.children)

    @property
    def tags(self) -> Set[Optional[str]]:
        """ Recursive access to tags on all leaf nodes """
        if self.is_branch:
            found_tags = set()
            for child in self.children:
                found_tags.update(child.tags)
            return found_tags

        # The getter can return a set containing None
        # while the setter only accepts a set containing strs.
        # mypy doesn't like this
        return self._tags or {None}  # type: ignore

    @tags.setter
    def tags(self, value: Set[str]) -> None:
        """ Recursive access to tags on all leaf nodes """
        if self.is_branch:
            for child in self.children:
                # see comment in getter
                child.tags = value  # type: ignore
        else:
            self._tags = value

    def is_idempotent_command(self, other_children: Iterable[HConfigChild]) -> bool:
        """ Determine if self.text is an idempotent change. """
        # Blacklist commands from matching as idempotent
        for rule in self.options["idempotent_commands_blacklist"]:
            if self.lineage_test(rule, True):
                return False

        # Handles idempotent acl entry identification
        if self._idempotent_acl_check():
            if self.host.os in {"iosxr"}:
                self_sn = self.text.split(" ", 1)[0]
                for other_child in other_children:
                    other_sn = other_child.text.split(" ", 1)[0]
                    if self_sn == other_sn:
                        return True

        # Idempotent command identification
        for rule in self.options["idempotent_commands"]:
            if self.lineage_test(rule, True):
                for other_child in other_children:
                    if other_child.lineage_test(rule, True):
                        return True

        return False

    def sectional_overwrite_no_negate_check(self) -> bool:
        """
        Check self's text to see if negation should be handled by
        overwriting the section without first negating it
        """
        for rule in self.options["sectional_overwrite_no_negate"]:
            if self.lineage_test(rule):
                return True
        return False

    def sectional_overwrite_check(self) -> bool:
        """ Determines if self.text matches a sectional overwrite rule """
        for rule in self.options["sectional_overwrite"]:
            if self.lineage_test(rule):
                return True
        return False

    def overwrite_with(
        self,
        other: HConfigChild,
        delta: Union[HConfig, HConfigChild],
        negate: bool = True,
    ) -> None:
        """ Deletes delta.child[self.text], adds a deep copy of self to delta """
        if other.children != self.children:
            if negate:
                delta.del_child_by_text(self.text)
                deleted = delta.add_child(self.text).negate()
                deleted.comments.add("dropping section")
            if self.children:
                delta.del_child_by_text(self.text)
                new_item = delta.add_deep_copy_of(self)
                new_item.comments.add("re-create section")

    def line_inclusion_test(
        self, include_tags: Set[str], exclude_tags: Set[str]
    ) -> bool:
        """
        Given the line_tags, include_tags, and exclude_tags,
        determine if the line should be included
        """
        include_line = False

        if include_tags:
            include_line = bool(self.tags.intersection(include_tags))
        if exclude_tags and (include_line or not include_tags):
            include_line = not bool(self.tags.intersection(exclude_tags))

        return include_line

    def lineage_test(self, rule: dict, strip_negation: bool = False) -> bool:
        """ A generic test against a lineage of HConfigChild objects """
        if rule.get("match_leaf", False):
            lineage_obj: Iterator[HConfigChild] = (o for o in (self,))
            lineage_depth = 1
        else:
            lineage_obj = self.lineage()
            lineage_depth = self.depth()

        rule_lineage_len = len(rule["lineage"])
        if rule_lineage_len != lineage_depth:
            return False

        matches = 0
        for lineage_rule, section in zip(rule["lineage"], lineage_obj):
            object_rules, text_match_rules = self._explode_lineage_rule(lineage_rule)

            if not self._lineage_eval_object_rules(object_rules, section):
                return False

            # This removes negations for each section but honestly,
            # we really only need to do this on the last one
            if strip_negation:
                if section.text.startswith(self._negation_prefix):
                    text = section.text[len(self._negation_prefix) :]
                elif section.text.startswith("default "):
                    text = section.text[8:]
                else:
                    text = section.text
            else:
                text = section.text

            if self._lineage_eval_text_match_rules(text_match_rules, text):
                matches += 1
                continue
            return False

        return matches == rule_lineage_len

    def _swap_negation(self) -> HConfigChild:
        """ Swap negation of a self.text """
        if self.text.startswith(self._negation_prefix):
            self.text = self.text[len(self._negation_prefix) :]
        else:
            self.text = self._negation_prefix + self.text

        return self

    def _default(self) -> HConfigChild:
        """ Default self.text """
        if self.text.startswith(self._negation_prefix):
            self.text = "default " + self.text[len(self._negation_prefix) :]
        else:
            self.text = "default " + self.text
        return self

    def _idempotent_acl_check(self) -> bool:
        """
        Handle conditional testing to determine if idempotent acl handling for iosxr should be used
        """
        if self.host.os in {"iosxr"}:
            if isinstance(self.parent, HConfigChild):
                acl = ("ipv4 access-list ", "ipv6 access-list ")
                if self.parent.text.startswith(acl):
                    return True
        return False

    @staticmethod
    def _explode_lineage_rule(rule: dict) -> Tuple[list, list]:
        text_match_rules: List[dict] = list()
        object_rules = list()
        for test, expression in rule.items():
            if test in {"new_in_config", "negative_intersection_tags"}:
                object_rules.append({"test": test, "expression": expression})
            elif test == "equals":
                if isinstance(expression, list):
                    text_match_rules.append(
                        {"test": test, "expression": set(expression)}
                    )
                else:
                    text_match_rules.append({"test": test, "expression": {expression}})
            elif test in {"startswith", "endswith"}:
                if isinstance(expression, list):
                    text_match_rules.append(
                        {"test": test, "expression": tuple(expression)}
                    )
                else:
                    text_match_rules.append({"test": test, "expression": (expression,)})
            elif isinstance(expression, list):
                text_match_rules += [
                    {"test": test, "expression": e} for e in expression
                ]
            else:
                text_match_rules += [{"test": test, "expression": expression}]
        return object_rules, text_match_rules

    def _lineage_eval_object_rules(self, rules: list, section: HConfigChild) -> bool:
        """
        Evaluate a list of lineage object rules.

        All object rules must match in order to return True

        """
        matches = 0
        for rule in rules:
            if rule["test"] == "new_in_config":
                if rule["expression"] == section.new_in_config:
                    matches += 1
                    continue
                return False
            if rule["test"] == "negative_intersection_tags":
                rule["expression"] = self._to_list(rule["expression"])
                if not set(rule["expression"]).intersection(section.tags):
                    matches += 1
                    continue
                return False
        return matches == len(rules)

    @staticmethod
    def _lineage_eval_text_match_rules(rules: list, text: str) -> bool:
        """
        Evaluate a list of lineage text_match rules.

        Only one text_match rule must match in order to return True
        """
        for rule in rules:
            if text_match.dict_call(rule["test"], text, rule["expression"]):
                return True
        return False

    @staticmethod
    def _to_list(obj: Union[list, object]) -> list:
        return obj if isinstance(obj, list) else [obj]

    @staticmethod
    def _to_set(items: Union[str, List[str], Set[str]]) -> Set[str]:
        # There's code out in the wild that passes List[str] or str, need to normalize for now
        if isinstance(items, list):
            return set(items)
        if isinstance(items, str):
            return {items}
        # Assume it's a set of str
        return items

    def _duplicate_child_allowed_check(self) -> bool:
        """ Determine if duplicate(identical text) children are allowed under the parent """
        for rule in self.options["parent_allows_duplicate_child"]:
            if self.lineage_test(rule):
                return True
        return False
